import 'dart:async';
import 'package:chemistry/src/atom.dart';
import 'package:chemistry/src/atom_does_not_exist_exception.dart';
import 'package:chemistry/src/chemistry.dart';

/// A Molecule is a derived state from one or multiple [Atom]s. This is a
/// generic class which allows it to be used very dynamically.
///
/// Example usage:
/// ```dart
/// var myMolecule = Molecule<int>(
///   atomKeys: ['someAtom'],
///   computer: (atoms) => atoms.values.fold<int>(
///     0, (previousValue, element) => previousValue + element.state),
///   key: 'myMolecule'));
/// myMolecule.stateStream.listen((state) => print(state));
/// ```
class Molecule<T> {
  /// A globally unique key. Usually the same as the name of this [Molecule].
  late final String key;

  /// The function used to compute the value of this [Molecule].
  late final T Function(Map<String, Atom<dynamic>> atoms) _computer;

  /// The keys of all atoms that should be used when calculating the value of
  /// this [Molecule].
  late final List<String> _atomKeys;

  /// Used to broadcast changes to this [Molecule]'s [_state] to all listeners.
  final StreamController<T> _controller = StreamController<T>.broadcast();

  /// This [Molecule]s state.
  late T _state;

  /// A broadcast stream that can be listened to. Whenever the state of this
  /// [Molecule] changes a new event will be emited on this stream providing the
  /// listener with the new state.
  Stream<T> get stateStream => _controller.stream;

  /// Get the state of this [Molecule].
  T get state => _state;

  /// Cosntructor.
  /// [atomKeys] the keys of the [Atom]s to be used when computing this
  /// [Molecule]'s value.
  /// [computer] the function that computes this [Molecule]'s value.
  /// [key] this [Molecule]'s globally unique key.
  ///
  /// Throws an AtomDoesNotExistException if an atom wasn't found in the
  /// chemistry store.
  ///
  /// Returns the newly created [Molecule].
  Molecule(
      {required List<String> atomKeys,
      required T Function(Map<String, Atom<dynamic>> atoms) computer,
      required String key})
      : this._atomKeys = atomKeys,
        this._computer = computer,
        this.key = key {
    _atomKeys.forEach((atomKey) {
      var atom = Chemistry().getAtom(atomKey);

      if (atom == null) {
        throw AtomDoesNotExistException(
            'No atom with the key $atomKey could be found in the Chemistry store.');
      }

      atom.stateStream.listen((value) {
        _computeValue();
      });
    });

    _computeValue();
  }

  /// The [dispose] method should always be called before this [Molecule] is
  /// destroyed. It will close [_controller]. Think of this as a destructor that
  /// you have to manually call.
  void dispose() {
    _controller.close();
  }

  /// Called when the value of an [Atom] that is referenced in [_atomKeys]
  /// changes. It computes the new state of this [Molecule].
  ///
  /// Throws an AtomDoesNotExistException if an atom wasn't found in the
  /// chemistry store.
  void _computeValue() {
    var atoms = Map<String, Atom<dynamic>>.fromIterable(_atomKeys,
        key: (atomKey) => atomKey,
        value: (atomKey) {
          Atom<dynamic>? atom = Chemistry().getAtom<dynamic>(atomKey);

          if (atom == null) {
            throw AtomDoesNotExistException(
                'No atom with the key $atomKey could be found in the Chemistry store.');
          }

          return atom;
        });
    _state = _computer(atoms);
    _controller.add(_state);
  }
}
